/*
 * Copyright 2011 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.robokind.api.animation.editor;

import org.robokind.api.common.playable.Playable;
import org.robokind.api.animation.editor.history.HistoryStack;
import org.robokind.api.common.utils.Source;
import org.robokind.api.animation.Animation;
import org.robokind.api.animation.Channel;
import org.robokind.api.common.config.VersionProperty;
import org.robokind.api.common.services.addon.ServiceAddOn;
import static org.robokind.api.common.localization.Localizer.$;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public class AnimationEditor extends AbstractEditor<Channel,ChannelEditor>{
    /**
     * Property String for the Animaiton's file path.
     */
    public final static String PROP_PATH = "Path";
    
    private Animation myAnimation;
    
    //TODO: Change this from a file path to a URI
    private String myPath;

    /**
     *
     * @param a
     * @param ps
     * @param path
     * @param hist
     */
    public AnimationEditor(Animation a, String path, HistoryStack hist){
        super(hist);
        myAnimation = a.clone();
        myPath = path;
        setChildren(myAnimation.getChannels());
    }
    
    /**
     *
     * @return
     */
    public String getFilePath(){
        return myPath;
    }

    /**
     *
     * @param path
     */
    public void setFilePath(String path){
        String oldPath = myPath;
        myPath = path;
        firePropertyChange(PROP_PATH, oldPath, myPath);
    }
    /**
     *
     * @param i
     * @param sel
     */
    @Override
    protected void setSelected(int i, boolean sel, HistoryStack hist){
        if(i < 0 || i >= myChildren.size()){
            return;
        }
        ChannelEditor child = getChild(i);
        if(child == null){
            return;
        }
        child.setState(this, EditState.SELECTED, sel, hist);
        if(!sel){
            return;
        }
        if(child.getSelectedIndex() == -1){
            if(child.getChildren().isEmpty()){
                return;
            }
            child.select(this, 0, hist);
        }
    }

    /**
     *
     * @return
     */
    @Override
    public String getName(){
        VersionProperty ver = myAnimation.getVersion();
        if(ver != null){
            return ver.display();
        }
        return $("new.animation");
    }
    
    /**
     *
     * @param name
     */
    @Override
    public void setName(String name){
        if(name == null || name.isEmpty()){
            return;
        }
        VersionProperty oldVersion = myAnimation.getVersion();
        String versionNumber = myAnimation.getVersion().getNumber();
        myAnimation.setVersion(name, versionNumber);
        firePropertyChange(PROP_NAME, oldVersion.display(), myAnimation.getVersion().display());
    }

    /**
     *
     * @param versionNumber
     */
    public void setVersionNumber(String versionNumber){
        VersionProperty oldVersion = myAnimation.getVersion();
        String name = myAnimation.getVersion().getName();
        myAnimation.setVersion(name, versionNumber);
        firePropertyChange(PROP_NAME, oldVersion.display(), myAnimation.getVersion().display());
    }
    public VersionProperty getVersion(){
        return myAnimation.getVersion();
    }
    /**
     *
     * @param version
     */
    public void setVersion(VersionProperty version){
        if(version == null){
            return;
        }
        VersionProperty oldVersion = myAnimation.getVersion();
        String name = version.getName();
        String ver = version.getNumber();
        if(name == null || name.isEmpty()){
            name = myAnimation.getVersion().getName();
        }
        myAnimation.setVersion(name, ver);
        firePropertyChange(PROP_NAME, oldVersion.display(), myAnimation.getVersion().display());
    }
    

    @Override
    protected Channel removeChild(Object invoker, int i){
        return myAnimation.removeChannelByListOrder(i);
    }

    /**
     *
     * @return
     */
    public Animation getAnimation(){
        return myAnimation.clone();
    }

    /**
     *
     * @return
     */
    public Animation getEnabledAnimation(){
        Animation anim = new Animation();
        for(ChannelEditor cc : myChildren){
            if(cc.hasFlag(EditState.DISABLED)){
                continue;
            }
            anim.addChannel(cc.getChannelView());
        }
        Animation a = anim.clone();
        for(ServiceAddOn<Playable> addon : myAnimation.getAddOns()){
            a.addAddOn(addon);
        }
        return a;
    }

    /**
     *
     * @param i
     * @return
     */
    public boolean containsLogicalId(int i){
        return myAnimation.containsLogicalId(i);
    }

    /**
     *
     * @return
     */
    public long getEnd(){
        long end = 0;
        for(ChannelEditor mpc : myChildren){
            long nEnd = mpc.getEnd();
            if(nEnd > end){
                end = nEnd;
            }
        }
        return end;
    }

    @Override
    protected ChannelEditor createChildController(Channel channel) {
        return new ChannelEditor(channel, mySharedHistory);
    }

    @Override
    public boolean isChildUIController() {
        return true;
    }

    @Override
    protected int addChildBase(Object invoker, Source<Channel> channelSource, int i) {
        myAnimation.insertChannel(i, channelSource.get());
        return myAnimation.getChannels().indexOf(channelSource.get());
    }

    @Override
    protected int insertChildControllerBase(ChannelEditor controller, int i){
        Channel channel = controller.getChannel();
        myAnimation.insertChannel(i, channel);
        return myAnimation.getChannels().indexOf(channel);
    }
}
